/*             CSE 687 Project 2 --- String Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * string.hh ---- header file for the module of the member & frind *
 *		  functions for class String and its derived class *
 * 		  BCstring  					   *
 *******************************************************************/

#ifndef String_HH

#define String_HH
#include <iostream.h>     // cin, cout, cerr
#include <stdlib.h>      // exit
#include <string.h>      // strlen, strcat, strcpy, strncpy 

// String class definition
class String {
   friend class BCstring;
   protected:   // for derived and friend classes
	int slen;
	char *shead;
	char *sptr;
   public:
	String(void);  // String s[10]  - define arrays of objects
	String(const String& s);  // String s2 = s1 - copy constructor
	String(const char *s);    // String s("...") - promotes a char* type
				  //   object to a String object
	String(const int n);      // String s(10) - constructs a blank String 
                                  // with only space allocated
	~String(void) { if (shead) delete []shead;}  // destructor 
	String& operator= (const String& str); // unary operator - s3 = s2 = s1
	// index operator (without bound checking)
	virtual char& operator[] (const int n) { return (*(sptr = shead + n));} 
	// relational friend operator	  
	friend int operator== (String& s1, String& s2) 
	  { return (strcmp(s1.shead, s2.shead) == 0); } 
	// binary operator + --- concatenation of two strings 
	// without bound checking
	friend String operator+ (String& s1, String& s2);
 	// overloading istream operator >>
	friend istream& operator>> (istream& istrm, String& str); 
 	// overloading ostream operator << 
	friend ostream& operator<< (ostream& ostrm, String& str) 
	  { return (ostrm << str.shead);} 
};	

// BCstring class (String class with bound checking) definition
class BCstring : public String {
   public:
	BCstring(void) :String(){ }  // BCstring s[10]  - define arrays of objects
	// BCstring s2 = s1 - copy constructor
	BCstring(const BCstring& s) : String(s){};  
	// BCstring s("...") - promotes a char* type object to a BCstring object
	BCstring(const char *s) : String(s){};
	// BCstring s(10) - constructs a blank String with only space allocated	  
	BCstring(const int n):String(n){};    
	// - promotes a String type object to a BCstring object
	BCstring(String& s) :String(s){this->slen = s.slen;}; 
	// index operator with bound checking
	char& operator[] (const int n); 
	// unary operator - s3 = s2 = s1 with bound checked
	BCstring& operator= (const BCstring& str); 
	BCstring& operator= (const String& str); 
	// binary operator + --- concatenation of two strings
	// with bound checking
	friend BCstring operator+ (BCstring& s1, BCstring& s2);
};
	
#endif String_HH		

/*             CSE 687 Project 2 --- String Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * string.cc ---- the module of the member & frind functions       *
 *                for class String and its derived class BCstring  *
 *******************************************************************/

//#include "string.hh"

extern void error(char *);      // error handler

///////////// class String member functions ///////////////

// ----<String :: String(void)>---- //
// - copy constructs String object (a null string)
String :: String(void)
{
  slen = 0;
  if ((shead = (sptr = new char[1])) == '\0') 
    error("error: exhaust system memory\n"); 
  *shead = '\0';
}

// ----<String :: String(const String& s)>---- //
// - copy constructs String object from an existing object
String :: String(const String& s)
{
  slen = s.slen;
  if ((shead = (sptr = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n"); 
  strcpy(shead, s.shead);
}

// ----<String :: String(const char *s)>---- //
// - constructs String object by promoting char* object
String :: String(const char *s)
{
  if (s == '\0') error("error: invalid string\n"); // check invalid string  
  slen = strlen(s);
  if ((shead = (sptr = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n"); 
  strcpy(shead, s);  
}

// ----<String :: String(const int n)>---- //
// - allocate space for a String with a length n and 
//   blank char is filled in the new string
String :: String(const int n)
{
  if (n<0) error("invalid string length\n");

  slen = n;
  if ((shead = (sptr = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n");
  for (int i = 0; i<slen; shead[i++] = ' ');
  shead[slen] = '\0';   
}

// ----<String& String :: operator= (const String& str)>---- //
// - sets String object equal to object str
String& String :: operator= (const String& str)
{ 
  if (*this == str) return *this; // for c1 = c1 case
  if (shead) delete []shead;
  slen = str.slen;
  if ((sptr = (shead = new char[slen+1])) == '\0') 
    error("error: exhaust system memory\n"); 
  strcpy(shead,str.shead);  // copy str's data
  
  return *this;
}

// ----<String operator+ (String& s1, String& s2)>---- //
// - concatenation of two strings
String operator+ (String& s1, String& s2)
{
  int length = s1.slen + s2.slen;
  char *tptr;
  if ((tptr = new char[length+1]) == '\0') 
      error("error: exhaust system memory\n"); 

  strcpy(tptr,s1.shead);    // copy s1's data
  strcat(tptr,s2.shead);    // copy s2's data
  String temp(tptr);  // use togather with unary operator = which
                      // copies this temporary immediately exit 
                      //from this function

  return temp;
}

// ----<istream& operator>> (istream& istrm, String& str)>---- //
// - overloading istream operator >>
istream& operator>> (istream& istrm, String& str)
{
  #define MAXSIZE 4095
  char *buffer;

  if ((buffer = new char[MAXSIZE+1]) == '\0') 
      error("error: exhaust system memory\n"); 
  istrm.getline(buffer, MAXSIZE);
  buffer[MAXSIZE] = '\0';
  str.slen = strlen(buffer);
  if (str.shead) delete str.shead;
  if ((str.shead = (str.sptr = new char[str.slen+1])) == '\0')
      error("error: exhaust system memory\n"); 
  strcpy (str.shead, buffer);
  delete []buffer;
  return istrm;

  #undef MAXSIZE
}

///////////// class BCstring member functions ///////////////

// ----<char& BCstring :: operator[] (const int n)>---- //
// - return ing reference allows writing and reading indexed data
char& BCstring :: operator[] (const int n)
{    
  if (n < 0 || n > strlen(shead)+1 ) {
    error ("error: index outside string\n");};
  sptr = shead + n;

  return shead[n];
}

// ----<BCstring& BCstring :: operator= (const BCstring& str)>---- //
// - sets String object equal to object str
BCstring& BCstring :: operator= (const BCstring& str)
{ 
  if (*this == str) return *this; // for c1 = c1 case

  // if source string has the same length as target string
  // directly perform copy
  if (slen != str.slen) { // not the same
    if (shead) delete []shead;
    slen = str.slen;
    if ((sptr = (shead = new char[slen+1])) == '\0') 
      error("error: exhaust system memory\n"); 
  }; // otherwise, use the previous space
  strcpy(shead,str.shead);  // copy str's data
  
  return *this;
}

// ----<BCstring& BCstring :: operator= (const String& str)>---- //
// - sets String object equal to object str
BCstring& BCstring :: operator= (const String& str)

{ 
  if (*this == str) return *this; // for c1 = c1 case

  // if source string has the same length as target string
  // directly perform copy
  if (slen != str.slen) { // not the same
    if (shead) delete []shead;
    slen = str.slen;
    if ((sptr = (shead = new char[slen+1])) == '\0') 
      error("error: exhaust system memory\n"); 
  }; // otherwise, use the previous space
  strcpy(shead,str.shead);  // copy str's data
  
  return *this;
}

// ----<BCstring operator+ (BCstring& s1, BCstring& s2)>---- //
// - concatenation of two strings
BCstring operator+ (BCstring& s1, BCstring& s2)
{
  int ltmp,length = strlen(s1.shead);

  if (s1.slen != length || s2.slen != (ltmp = strlen(s2.shead))) {
    cerr << "warning: incorrect string length\n";
    length += ltmp;
  }
  else 
    length = s1.slen + s2.slen;

  char *tptr;
  if ((tptr = new char[length+1]) == '\0') 
      error("error: exhaust system memory\n"); 

  strcpy(tptr,s1.shead);    // copy s1's data
  strcat(tptr,s2.shead);    // copy s2's data
  String temp(tptr);  // use togather with unary operator = which
                      // copies this temporary immediately exit 
                      //from this function

  return temp;
}


/*             CSE 687 Project 2 --- String Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * proj2.cc ---- main program to test the member & frind functions *
 *               for class String and its derived class BCstring   *
 *******************************************************************/

#include <stdio.h>  // NULL
#include <fstream.h> 
//#include "string.hh"  // class String

extern void error(char *);      // error 

/* function prototypes */
void compact_line(String& sline,BCstring& cline);
int function_line(BCstring& sline,const int i);

//--- <void error(char *s)> ---//
/* error handler */
void error(char *s)
{
  cerr << s; 
  exit(1);
}

//--- <main program> ---//
/* test the member and friend functions for class String
   and its derived class BCstring */
main(int argc, char *argv[])
{
  // check input file on command line
  if (argc < 2) error("Please enter file name on command line\n");

  ifstream fp(argv[1]);
 
  // check i/o status
  if (!fp) {
    cerr << "cannot open file: " << argv[1] << "\n";
    exit(1);
  };
  // status information     
  cout << "This is file " << argv[1]
    << "\nProcessed by " <<  argv[0] << "\n\n";

  String sline; // input line buffer.(To illustrate
                // the use of String class and BCstring class, we
                // use String here to store the original line)

  long lines = 0; // line counter
  while (fp) {
    fp >>sline; // read in a line
    lines++;

    int i = 0; char ch;

    // get ride of comment part of a line
    while ((ch = sline[i++])!= '/' && (ch != '\0'));
    if (ch = '/') sline[--i] = '\0';

    // find out if the line contains '(' (possible wanted line)
    i = 0;
    while ((ch = sline[i++])!= '(' && (ch != '\0'));
 
    if (ch == '(') {   // line containing '('

      BCstring cline;  // compacted line buffer.(To illustrate
                       // the use of String class and BCstring class, we
                       // use BCstring here to store the compacted line)

      compact_line(sline,cline); // get ride of blank space,
                                 // compacted line is in cline 
      // relocate '(' in the compacted line and test if the line
      // contains function declaration, definition, invocation etc.
      i = 0;
      char lines_string[15];
      do {
	while ( (ch = cline[i]) != '(' && (ch != '\0')) i++;
	if ((ch == '(') && function_line(cline,i)) { // find wanted line
	  sprintf(lines_string,"Line %5d: ",lines);
	  // print out the line
	  cout <<(String&)lines_string + sline << "\n"; 
	  // or cout <<(String&)lines_string + sline + (String&)"\n"
	  break;
	};
	i++;
      } while (ch != '\0');
    };
  };
      
  exit(0); 
}

//---<void compact_line(BCstring& sline,BCstring& cline)>---//
/* get ride of blank space in a line */
void compact_line(String& sline,BCstring& cline) // get ride of blank space 
{
  int i = 0, j = 0;
  char ch;
  cline = sline;  // copy a line

  while ((ch = sline[i++]) != '\0')  // skip blank space
    if (ch != ' ' && ch != '\t') cline[j++] = ch; // copy non-blank char.

  cline[j] = '\0';  // string terminal symbol

  return;
}

//---<int function_line(BCstring& sline, const int n)>---//
/* test if a line contains function declaration, definition, invocation etc. */
int function_line(BCstring& sline, const int n)
{
  // macro difinitions for test different '(' 
  // occurrences in expressions and statments
#define char_keyword(c) (c=='*'||c=='('||c==')'||c=='['|| c==']'||c==':'||c=='&'||\
	       	    c=='|'||c=='~'||c=='"'||c=='\''|| c=='!'||c=='%'||c=='>'||\
                    c=='<'||c=='+'||c=='-'||c=='/'|| c=='?'|| c=='='|| c==';') 

#define keyword(x) ((x) == "  case" || (x) == "    do" || (x) == "  else" ||\
                      (x) == "   for" || (x) == "  goto" || (x) == "    if" ||\
                      (x) == "return" || (x) == "sizeof" || (x) == "switch" ||\
                      (x) == " while" )
  #define true 1
  #define false 0  

  // first exclude possible occurring of '(' in an expression 
  // n is the location of '(' in a string
  if (n == 0 || ((n >0) && char_keyword(sline[n-1]))) return false;

  // exlude statements
  BCstring temp(6); // longest keyword such as "switch"
  int i, j;

  temp[6] = '\0'; 
  // exclude possible occurring of '(' in a statement
  for (i = 5, j = n - 1; i>=0; i--, j--) {     
     if (j>=0) temp[i] = sline[j];
     else temp[i] = ' ';
     if (keyword(temp)) return false; // assuming unfilled positions in
                                      // BCstring temp is ' '
  };

  return true;
}
